/*
 * Copyright (C) 2013 - Juan Ferrer Toribio
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include <libpq-fe.h>

#include "squid-db-logger.h"

static int exit_signal = FALSE;

typedef struct
{
	PGconn * db;
	int sock;
	int error;
	char * host;
	char * schema;
	char * user;
	char * password;
	int nparams;
	int port;
	char query_buffer[BUFFER_SIZE];
	FILE * log_file;
	size_t log_size;
}
Logger;

void logger_free (Logger * self)
{
	if (self->log_file)
		fclose (self->log_file);
	if (self->db)
		PQfinish (self->db);
	if (self->sock)
		shutdown (self->sock, 2);

	free (self->host);
	free (self->schema);
	free (self->user);
	free (self->password);
	free (self);
}

void logger_init_log (Logger * self)
{
	self->log_file = fopen (LOG_FILE, "a");

	if (self->log_file)
	{
		fseek (self->log_file, 0L, SEEK_END);
		self->log_size = ftell (self->log_file);
	}
	else
		self->log_file = stdout;
}

void logger_vlog (Logger * self, char * str, va_list vl)
{
	time_t now;
	struct tm * timeinfo;
	char time_buffer[80];

	now = time (NULL);
	timeinfo = localtime (&now);
	strftime (time_buffer, 80, "%Y-%m-%d %H:%M:%S", timeinfo);
	
	fprintf (self->log_file, "%s [%d]: ", time_buffer, getpid ());
	self->log_size += vfprintf (self->log_file, str, vl);
	fprintf (self->log_file, "\n");

	if (self->log_size > MAX_LOG_SIZE && self->log_file != stdout)
	{
		fclose (self->log_file);
		rename (LOG_FILE, LOG_FILE ".1");
		logger_init_log (self);
	}
}

void logger_log (Logger * self, char * str, ...)
{
	va_list vl;
	va_start (vl, str);
	logger_vlog (self, str, vl);
	va_end (vl);
}

void logger_exit (Logger * self, char * str, ...)
{
	va_list vl;
	va_start (vl, str);
	logger_vlog (self, str, vl);
	va_end (vl);

	logger_free (self);
	exit (EXIT_FAILURE);
}

int logger_init_db (Logger * self)
{
	int ok = TRUE;
	PGresult * res;

	if (self->db)
		PQfinish (self->db);

	self->db = PQsetdbLogin (
		 self->host
		,NULL
		,NULL
		,NULL
		,self->schema
		,self->user
		,self->password
	);

	ok = PQstatus (self->db) == CONNECTION_OK;

	if (ok)
	{
		res = PQprepare (self->db, "proxy_stmt", self->query_buffer, 0, NULL);
		ok = PQresultStatus (res) == PGRES_COMMAND_OK;
		PQclear (res);
	}		

	if (!ok)
		logger_log (self, PQerrorMessage (self->db));
	else
		logger_log (self, "Connection to database established.");

	return ok;
}

char * strtrim (char * str)
{
	char * end;
	
	if (!str)
		return NULL;

	while (isspace (*str))
		str++;

	end = str + strlen (str) - 1;

	while (end > str && isspace (*end))
		end--;

	*(end + 1) = 0;
	return str;
}

void signal_handler (int sig)
{
	exit_signal = TRUE;
}

int main (int argc, char * argv[])
{
	char buffer[BUFFER_SIZE];

	Logger * self = malloc (sizeof (Logger));
	self->db = NULL;
	self->sock = 0;
	self->host = NULL;
	self->schema = NULL;
	self->user = NULL;
	self->password = NULL;
	self->nparams = 0;
	self->port = 0;
	self->log_file = NULL;
	self->log_size = 0;

	logger_init_log (self);
	logger_log (self, "Daemon started.");

	// Reading the main config file

	char * key;
	char * value;
	FILE * file = fopen (CONFIG_FILE, "r");

	if (file)
	{
		while (fgets (buffer, BUFFER_SIZE - 1, file))
		{
			size_t len = strlen (buffer) - 1;

			if (len <= 0 || buffer[0] == '#')
				continue;
			
			key = strtrim (strtok (buffer, "="));

			if (!key)
				continue;

			value = strdup (strtrim (strtok (NULL, "\n")));

			if (!strcmp (key, "host"))
				self->host = value;
			else if (!strcmp (key, "schema"))
				self->schema = value;
			else if (!strcmp (key, "user"))
				self->user = value;
			else if (!strcmp (key, "password"))
				self->password = value;
			else if (!strcmp (key, "nparams"))
				self->nparams = atoi (value);
			else if (!strcmp (key, "port"))
				self->port = atoi (value);
			else
				free (value);
		}

		fclose (file);
	}
	else
		logger_exit (self, "Can't open the config file: %s", CONFIG_FILE);

	if (self->nparams <= 0)
		logger_exit (self, "The value of 'nparams' must be greather than 0.");
	if (!(self->port >= 0 && self->port <= 65535))
		logger_exit (self, "Invalid port number.");

	// Creating the socket

	struct sockaddr_in server_addr;

	self->sock = socket (AF_INET, SOCK_DGRAM, 0);

	if (self->sock != -1)
	{
		server_addr.sin_family = AF_INET;
		server_addr.sin_port = htons (self->port);
		server_addr.sin_addr.s_addr = htonl (INADDR_ANY);

		if (bind (self->sock, (struct sockaddr *) &server_addr, sizeof (server_addr)) == -1)
			logger_exit (self, "Can't bind on localhost:%d.", self->port);
	}
	else
		logger_exit (self, "Can't create the socket.");

	// Reading the insert query from the config file

	file = fopen (SQL_FILE, "r");

	if (file)
	{
		size_t len = fread (self->query_buffer, 1, BUFFER_SIZE - 1, file);
		fclose (file);

		if (len > 0)
			self->query_buffer[len] = '\0';
		else
			logger_exit (self, "Error reading the query file: %s", SQL_FILE);
	}
	else
		logger_exit (self, "Can't open the query file: %s", SQL_FILE);

	// Connecting to the database

	logger_log (self, "Connecting to the database.");
	logger_init_db (self);

	// Handling the SIGTERM signal

	struct sigaction action;

	memset (&action, 0, sizeof (struct sigaction));
	action.sa_handler = signal_handler;
	sigaction (SIGTERM, &action, NULL);
	sigaction (SIGHUP, &action, NULL);
	sigaction (SIGINT, &action, NULL);

	// Main loop

	int i;
	ssize_t len;
	PGresult * res;
	socklen_t client_len;
	struct sockaddr_in client_addr;
	const char * paramValues[self->nparams];
	char * next;

	while (!exit_signal)
	{
		client_len = sizeof (client_addr);
		
		len = recvfrom (self->sock, buffer, BUFFER_SIZE, 0,
			(struct sockaddr *) &client_addr, &client_len);

		if (len == -1)
			continue;

		buffer[len] = '\0';
		next = buffer;

		logger_log (self, "%s:%u: %s"
			,inet_ntoa (client_addr.sin_addr)
			,ntohs (client_addr.sin_port)
			,buffer
		);

		for (i = 0; i < self->nparams; i++)
			paramValues[i] = strtok_r (next, " \t", &next);
			
		res = PQexecPrepared (self->db, "proxy_stmt",
			self->nparams, paramValues, NULL, NULL, 0);
		
		if (PQresultStatus (res) != PGRES_COMMAND_OK)
		{
			logger_log (self, PQerrorMessage (self->db));

			if (PQstatus (self->db) == CONNECTION_BAD)
			{
				logger_log (self, "Connecting to database lost, trying to reconnect.");

				while (!logger_init_db (self) && !exit_signal)
				{
					logger_log (self, "Trying again in %d seconds.", RETRY_TIMEOUT);
					sleep (RETRY_TIMEOUT);
				}
			}
		}

		PQclear (res);
	}

	logger_log (self, "Daemon stopped.");
	logger_free (self);
	return EXIT_SUCCESS;
}
